/*
 * $Id: tfm.c,v 1.5 1992/01/25 22:10:26 trevor Exp $
 */

/*
 * Trevor's Implementation of the SGI Font Manager for use on 
 * machines that do not have the Font Manager (ie IBM)
 */

 /* Fonts were generate from and SGI running FM and converted to tfm
  * (Trev's Font Manager) using fm2tfm.c
  * because of a gl limitation fonts are limited to 256x256 pixels
  */

#include <stdio.h>
#include <unistd.h>
#include "tfmclient.h"
#include "tfminternal.h"

tfm_struct tfm;
tfm_font *tfmtmp, *tfmtmp2, *tfmcurr;
long fmwin;

extern tfm_font *tfm_gentfmfont();

void fminit()
{
 FILE *fp;
 char *getenv(), file[256], *f;
 static int tfminit = 0;
 int flen;

 if(tfminit)
   return;

 tfminit = 1;
 file[0] = 0;

 f = getenv("FONTPATH");
 if(f) {
   strcpy(file,f);
   flen = strlen(file) + 1;
   tfm.fontpath = (char *) malloc(flen);
   if(!tfm.fontpath) {
     perror("TFM");
     exit(-1);
   }
   strcpy(tfm.fontpath, file);
 }
 else {
   flen = strlen("/usr/local/lib/tfmfonts") + 1;
   tfm.fontpath = (char *) malloc(flen);
   if(!tfm.fontpath) {
     perror("TFM");
     exit(-1);
   }
   strcpy(tfm.fontpath, "/usr/local/lib/tfmfonts");
 }

 sprintf(file,"%s/tfmlist",tfm.fontpath);
 if(access(file,R_OK)) {
   perror(tfm.tfminfo);
   exit(-1);
 }
 flen = strlen(file) + 1;
 tfm.tfminfo = (char *)malloc(flen);
 if(!tfm.tfminfo) {
   perror("TFM");
   exit(-1);
 }

 strcpy(tfm.tfminfo,file);
 tfm.numfonts = 0;
 tfm.head = NULL;
 tfm.current = NULL;
}

void fmenumerate(callback)
void (*callback)();
{
 FILE *fp;
 unsigned char size;
 char fontname[256];
 long offset;

 fp = fopen(tfm.tfminfo, "r");
 if(!fp) {
   perror(tfm.tfminfo);
   exit(-1);
 }

 fontname[0] = 0;
 fscanf(fp,"%s ", fontname);
 while(!feof(fp)) {
   do {
     fread(&size, sizeof(unsigned char), 1, fp);
     fread(&offset, sizeof(long), 1, fp);
   }
   while(size);
   (callback)(fontname);
   fontname[0] = 0;
   fscanf(fp,"%s ", fontname);
 }
 fclose(fp);
}

fmfonthandle fmfindfont(face)
char *face;
{
 FILE *fp;
 long found;
 float s;
 char fontname[256];
 unsigned char size;
 long offset;

 found = (long)tfm_isloaded(face,1.0);
 if(found) {
   tfm_incref((fmfonthandle)found);
   return((fmfonthandle)found);
 }

 fp = fopen(tfm.tfminfo, "r");
 if(!fp) {
   perror(tfm.tfminfo);
   exit(-1);
 }
 fontname[0] = 0;
 fscanf(fp,"%s ", fontname);
 while(!feof(fp) && !found) {
   if(!strcmp(fontname, face))
     found = 1;
   else {
     do {
       fread(&size, sizeof(unsigned char), 1, fp);
       fread(&offset, sizeof(long), 1, fp);
     }
     while(size);
     fscanf(fp,"%s ", fontname);
   }
 }
 fclose(fp);
 if(!found) {
   fprintf(stderr,"TFM: Font %s not found\n", face);
   return(0);
 }
 /* font found, add it to the list of fonts */
 tfmtmp = tfm_gentfmfont();
 tfmcurr = tfm.head;
 if(tfmcurr == NULL) { /* first font in the list */
   tfm.head = tfmtmp;
 }
 else { /* add it to the end of the list */
   while(tfmcurr->next != NULL)
     tfmcurr = tfmcurr -> next;
   tfmcurr -> next = tfmtmp;
 }
 s = 1.0;
 offset = 0;
 /* load the font with the scale factor passed */
 tfm_loadfont(fontname, &offset, &s, tfmtmp);
 defrasterfont(tfmtmp->fontnum, tfmtmp->h, 128, tfmtmp->chars,
	       tfmtmp->numshorts, tfmtmp->rasters);
 tfmtmp -> fonthandle = (long)tfmtmp -> fontnum;
 tfmtmp -> realfontsize = 1.0;
 tfmtmp -> references++;
 return((fmfonthandle)&tfmtmp -> fonthandle);
}

fmfonthandle fmscalefont(tfh,scale)
fmfonthandle tfh;
double scale;
{
 long found = 0;
 float s,rs;
 long offset;
 long fh;

 fh = *tfh;
 tfmtmp = tfm.head;
 while(tfmtmp && !found) {
   if(tfmtmp->fontnum == (short)fh) {
     tfmcurr = tfmtmp;
     found = 1;
   }
   else 
     tfmtmp = tfmtmp -> next;
 }
 if(!found) {
   fprintf(stderr,"TFM: fmscalefont: font %ld not found\n",fh);
   return(0);
 }
 
 s = tfmcurr->fontsize * scale; /* get scale the user wants */
 rs = s; /* set REAL scale value */
 tfm_getscale(tfmcurr->fontname, &s, &offset); /* figure out what we can give */
 found = (long)tfm_isloaded(tfmcurr->fontname, rs); /* is the font loaded */
 if(found) {
   tfm_incref((fmfonthandle)found);
   return((fmfonthandle)found);
 }

 /* font found, add it to the list of fonts */
 tfmtmp = tfm_gentfmfont();
 tfmtmp2 = tfm.head;
 if(tfmtmp2 == NULL) { /* first font in the list */
   tfm.head = tfmtmp;
 }
 else { /* add it to the end of the list */
   while(tfmtmp2->next != NULL)
     tfmtmp2 = tfmtmp2 -> next;
   tfmtmp2 -> next = tfmtmp;
 }
 tfm_loadfont(tfmcurr->fontname, &offset, &s, tfmtmp);
 defrasterfont(tfmtmp->fontnum, tfmtmp->h, 128, tfmtmp->chars,
	       tfmtmp->numshorts, tfmtmp->rasters);
 tfmtmp -> fonthandle = (long)tfmtmp -> fontnum;
 tfmtmp -> realfontsize = rs;
 tfmtmp -> references++;
 return((fmfonthandle)&tfmtmp -> fonthandle);
}

void fmsetfont(tfh)
fmfonthandle tfh;
{
 long fh;

 fh = *tfh;
 tfmtmp = tfm.head;
 while(tfmtmp) {
   if(tfmtmp->fontnum == (short)fh) {
     if(tfm.current == tfmtmp) {
       /* already current font.. no need for font() overhead */
       return;
     }
     tfm.current = tfmtmp;
     font(tfmtmp->fontnum);
     return;
   }
   tfmtmp = tfmtmp -> next;
 }
 fprintf(stderr,"TFM: fmsetfont: font %ld not found\n",fh);
}

char *fmfontpath()
{
 return(tfm.fontpath);
}

void fmfreefont(tfh)
fmfonthandle tfh;
{
 long fh;

 fh = *tfh;
 tfmtmp = tfm.head;
 tfmtmp2 = tfmtmp;
 while(tfmtmp) {
   if(tfmtmp->fontnum == (short)fh) {
     tfmtmp->references--;
     if(tfmtmp->references == 0) {
       tfmtmp2 -> next = tfmtmp-> next;
       if(tfmtmp == tfm.current)
         tfm.current = NULL;
       if(tfmtmp == tfm.head)
         tfm.head = tfmtmp->next;
       tfm_freetfmfont(tfmtmp);
     }
     return;
   }
   tfmtmp2 = tfmtmp;
   tfmtmp = tfmtmp -> next;
 }
 fprintf(stderr,"TFM: fmfreefont: font %ld not found\n",fh);
 return;
}

long fmgetchrwidth(tfh, ch)
fmfonthandle tfh;
unsigned char ch;
{
 long fh;

 fh = *tfh;
 tfmtmp = tfm.head;
 tfmtmp2 = tfmtmp;
 while(tfmtmp) {
   if(tfmtmp->fontnum == (short)fh) {
     if((ch < 0x20) || (ch > 127))
       ch = 0x20; /* make ch a space */
     return((long)tfmtmp->chars[ch].width);
   }
   tfmtmp = tfmtmp -> next;
 }
 fprintf(stderr,"TFM: fmgetchrwidth: font %ld not found\n",fh);
 return((long)-1);
}

int fmgetfontinfo(tfh, info)
fmfonthandle tfh;
fmfontinfo *info;
{
 long fh;

 fh = *tfh;
 tfmtmp = tfm.head;
 tfmtmp2 = tfmtmp;
 while(tfmtmp) {
   if(tfmtmp->fontnum == (short)fh) {
     memcpy(info, &tfmtmp->fontinfo, sizeof(fmfontinfo));
     return(1);
   }
   tfmtmp = tfmtmp -> next;
 }
 fprintf(stderr,"TFM: fmgetfontinfo: font %ld not found\n",fh);
 return(-1);
}


int fmgetfontname(tfh, slen, str)
fmfonthandle tfh;
int slen;
char *str;
{
 long fh;

 fh = *tfh;
 tfmtmp = tfm.head;
 tfmtmp2 = tfmtmp;
 while(tfmtmp) {
   if(tfmtmp->fontnum == (short)fh) {
     if(strlen(tfmtmp->fontname) < slen)
       slen = strlen(tfmtmp->fontname);
     strncpy(str,tfmtmp->fontname,slen);
     return(strlen(str));
   }
   tfmtmp = tfmtmp -> next;
 }
 fprintf(stderr,"TFM: fmgetfontinfo: font %ld not found\n",fh);
 return(-1);
}


long fmgetstrwidth(tfh, str)
fmfonthandle tfh;
char *str;
{
 long fh;
 long ret;

 fh = *tfh;
 tfmtmp = tfm.head;
 tfmtmp2 = tfmtmp;
 while(tfmtmp) {
   if(tfmtmp->fontnum == (short)fh) {
     font(tfmtmp->fontnum);
     ret = strwidth(str);
     if(tfm.current)
       font(tfm.current->fontnum);
     return(ret);
   }
   tfmtmp = tfmtmp -> next;
 }
 fprintf(stderr,"TFM: fmgetstrwidth: font %ld not found\n",fh);
 return(-1);
}

int fmgetwholemetrics(tfh,gi)
fmfonthandle tfh;
fmglyphinfo *gi;
{
 long fh;

 fh = *tfh;
 tfmtmp = tfm.head;
 tfmtmp2 = tfmtmp;
 while(tfmtmp) {
   if(tfmtmp->fontnum == (short)fh) {
     memcpy(gi,tfmtmp->fmchars,sizeof(fmglyphinfo)*128);
     return(0);
   }
   tfmtmp = tfmtmp -> next;
 }
 fprintf(stderr,"TFM: fmgetstrwidth: font %ld not found\n",fh);
 return(-1);
}

long fmoutchar(tfh, ch)
fmfonthandle tfh;
unsigned char ch;
{
 unsigned char st[2];
 long fh;

 fh = *tfh;
 tfmtmp = tfm.head;
 tfmtmp2 = tfmtmp;
 while(tfmtmp) {
   if(tfmtmp->fontnum == (short)fh) {
     if((ch < 0x20) || (ch > 127))
       ch = (unsigned char)0x20; /* make ch a space */
     st[0] = ch;
     st[1] = 0;
     charstr(st);
     return((long)tfmtmp->chars[ch].width);
   }
   tfmtmp = tfmtmp -> next;
 }
 fprintf(stderr,"TFM: fmoutchar: font %ld not found\n",fh);
 return(-1);
}

int fmprstr(str)
char *str;
{
 if(str == NULL)
   return(-1);

 if(tfm.current) {
   charstr(str);
   return((long)strwidth(str));
 }
 fprintf(stderr,"TFM: fmprstr: font not set. Use fmsetfont()\n");
 return(-1);
}

void fmsetpath(path)
char *path;
{
 char file[256];
 FILE *fp;

 if(tfm.fontpath)
   free(tfm.fontpath);
 tfm.fontpath = (char *) malloc(strlen(path+1));
 strcpy(tfm.fontpath, path);

 sprintf(file,"%s/tfmlist",tfm.fontpath);
 if(access(file,R_OK)) {
   perror(tfm.tfminfo);
   exit(-1);
 }
 if(tfm.tfminfo)
   free(tfm.tfminfo);
 tfm.tfminfo = (char *)malloc(strlen(file)+1);
 strcpy(tfm.tfminfo,file);
}
/* ---------------------------------------------------------------------- */
/* Unused/unimplemented functions */
void            fmcachedisable(){}

void            fmcacheenable(){}

int 		fmcachelimit(){ return(0);}

int             fmgetcacheused(){return(0);}

int fmgetcomment(tfh, slen, str)
fmfonthandle tfh;
int slen;
char *str;
{ return(-1);}

fmfonthandle fmmakefont(tfh, m)
fmfonthandle tfh;
double m[3][2];
{return(0);}

void            fminitpagematrix(){}
void            fmsetpagematrix(m)double m[3][2];{}
void            fmgetpagematrix(m)double m[3][2];{}
void            fmscalepagematrix(m)double m;{}
void            fmrotatepagematrix(m)double m;{}
void            fmconcatpagematrix(m)double m[3][2];{}

void            fmprintermatch(set)int set;{}

void            fmsetcachelimit(new_limit)long new_limit;{}

/*
 * $Log: tfm.c,v $
 * Revision 1.5  1992/01/25  22:10:26  trevor
 * - changed error message for fmgetchrwidth
 * - in fmgetstrwidth only reset the font if the font has been set
 *
 * Revision 1.4  1991/10/10  22:47:43  trevor
 * bugfix for IBM
 *
 * Revision 1.3  1991/09/23  18:41:25  trevor
 * - New (Compacted) font file format
 * - many bug fixes
 *
 * Revision 1.2  1991/07/09  16:55:30  trevor
 * fixed fmfreefont bug, not resetting pointers properly when deleting a
 * font. This caused the "mdl menu with no characters" bug.
 *
 * Revision 1.1  1991/07/04  21:23:13  trevor
 * only call tfm.init if never called before (duh..)
 *
 * Revision 1.0  1991/06/28  22:52:48  trevor
 * Initial checkin
 *
 */
